@@ -527,7 +527,8 @@ gnu_write (PedDevice* dev, const void* buffer, PedSector start, PedSector count)
 	   store block, then we need to fetch the old block, copy in the
 	   overlaping area and finally, write the modified data out to the
 	   store.  */
-	if ((PED_SECTOR_SIZE_DEFAULT * start) % arch_specific->store->block_size != 0) {
+	if ((PED_SECTOR_SIZE_DEFAULT * start) % arch_specific->store->block_size
+	    != 0) {
 		size_t 		write_offset;
 		size_t 		flushing;
 
@@ -619,7 +620,7 @@ doggy_first_block_write:
 	}
 
 	while (count > 0
-	       && count >= arch_specific->store->block_size / PED_SECTOR_SIZE_DEFAULT) {
+	   && count >= arch_specific->store->block_size / PED_SECTOR_SIZE_DEFAULT) {
 		err = store_write (arch_specific->store,
 				   PED_TO_STORE (arch_specific->store, start),
 				   buffer, count * PED_SECTOR_SIZE_DEFAULT,
