@@ -17,6 +17,12 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */
 
+/* we need this for O_DIRECT on Linux, it's in the glibc headers */
+/* ... and for posix_memalign()                                  */
+#define _GNU_SOURCE
+
+#define PROC_DEVICES_BUFSIZ 16384
+
 #include "config.h"
 
 #include <parted/parted.h>
@@ -29,6 +35,7 @@
 #include <libgen.h>
 #include <stdint.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <syscall.h>
 #include <unistd.h>
@@ -38,6 +45,9 @@
 #include <sys/types.h>
 #include <sys/utsname.h>        /* for uname() */
 #include <scsi/scsi.h>
+#ifdef ENABLE_DEVICE_MAPPER
+#include <libdevmapper.h>
+#endif
 
 #include "blkpg.h"
 
@@ -62,6 +72,19 @@
 #define HDIO_GETGEO             0x0301  /* get device geometry */
 #define HDIO_GET_IDENTITY       0x030d  /* get IDE identification info */
 
+#include <parted/vtoc.h>
+#include <parted/fdasd.h>
+
+#if defined(O_DIRECT) && (!defined(__s390__) || !defined(__s390x__))
+#define RD_MODE (O_RDONLY | O_DIRECT)
+#define WR_MODE (O_WRONLY | O_DIRECT)
+#define RW_MODE (O_RDWR | O_DIRECT)
+#else
+#define RD_MODE (O_RDONLY)
+#define WR_MODE (O_WRONLY)
+#define RW_MODE (O_RDWR)
+#endif
+
 struct hd_geometry {
         unsigned char heads;
         unsigned char sectors;
@@ -90,15 +113,18 @@ struct hd_driveid {
         unsigned short  vendor2;        /* vendor unique */
         unsigned char   serial_no[20];  /* 0 = not_specified */
         unsigned short  buf_type;
-        unsigned short  buf_size;       /* 512 byte increments; 0 = not_specified */
-        unsigned short  ecc_bytes;      /* for r/w long cmds; 0 = not_specified */
+        unsigned short  buf_size;       /* 512 byte increments;
+                                                         0 = not_specified */
+        unsigned short  ecc_bytes;      /* for r/w long cmds;
+                                                         0 = not_specified */
         unsigned char   fw_rev[8];      /* 0 = not_specified */
         char            model[40];      /* 0 = not_specified */
         unsigned char   max_multsect;   /* 0=not_implemented */
         unsigned char   vendor3;        /* vendor unique */
         unsigned short  dword_io;       /* 0=not_implemented; 1=implemented */
         unsigned char   vendor4;        /* vendor unique */
-        unsigned char   capability;     /* bits 0:DMA 1:LBA 2:IORDYsw 3:IORDYsup*/
+        unsigned char   capability;     /* bits 0:DMA 1:LBA 2:IORDYsw
+                                                3:IORDYsup*/
         unsigned short  reserved50;     /* reserved (word 50) */
         unsigned char   vendor5;        /* vendor unique */
         unsigned char   tPIO;           /* 0=slow, 1=medium, 2=fast */
@@ -117,7 +143,8 @@ struct hd_driveid {
         unsigned short  dma_mword;      /* multiple-word dma info */
         unsigned short  eide_pio_modes; /* bits 0:mode3 1:mode4 */
         unsigned short  eide_dma_min;   /* min mword dma cycle time (ns) */
-        unsigned short  eide_dma_time;  /* recommended mword dma cycle time (ns) */
+        unsigned short  eide_dma_time;  /* recommended mword dma cycle
+                                           time (ns) */
         unsigned short  eide_pio;       /* min cycle time (ns), no IORDY  */
         unsigned short  eide_pio_iordy; /* min cycle time (ns), with IORDY */
         unsigned short  words69_70[2];  /* reserved words 69-70 */
@@ -127,9 +154,11 @@ struct hd_driveid {
         unsigned short  words76_79[4];  /* reserved words 76-79 */
         unsigned short  major_rev_num;  /*  */
         unsigned short  minor_rev_num;  /*  */
-        unsigned short  command_set_1;  /* bits 0:Smart 1:Security 2:Removable 3:PM */
+        unsigned short  command_set_1;  /* bits 0:Smart 1:Security 2:Removable
+                                                3:PM */
         unsigned short  command_set_2;  /* bits 14:Smart Enabled 13:0 zero */
-        unsigned short  cfsse;          /* command set-feature supported extensions */
+        unsigned short  cfsse;          /* command set-feature supported
+                                           extensions */
         unsigned short  cfs_enable_1;   /* command set-feature enabled */
         unsigned short  cfs_enable_2;   /* command set-feature enabled */
         unsigned short  csf_default;    /* command set-feature default */
@@ -178,7 +207,9 @@ struct hd_driveid {
 #define BLKSSZGET  _IO(0x12,104) /* get block device sector size */
 #define BLKGETLASTSECT  _IO(0x12,108) /* get last sector of block device */
 #define BLKSETLASTSECT  _IO(0x12,109) /* set last sector of block device */
-#define BLKGETSIZE64 _IOR(0x12,114,size_t)      /* return device size in bytes (u64 *arg) */
+
+/* return device size in bytes (u64 *arg) */
+#define BLKGETSIZE64 _IOR(0x12,114,size_t)
 
 struct blkdev_ioctl_param {
         unsigned int block;
@@ -229,6 +260,10 @@ struct blkdev_ioctl_param {
 #define I2O_MAJOR7              86
 #define I2O_MAJOR8              87
 #define UBD_MAJOR               98
+#define DASD_MAJOR              94
+#define VIODASD_MAJOR           112
+#define SX8_MAJOR1              160
+#define SX8_MAJOR2              161
 
 #define SCSI_BLK_MAJOR(M) (                                             \
                 (M) == SCSI_DISK0_MAJOR                                 \
@@ -268,6 +303,106 @@ _is_i2o_major (int major)
         return (I2O_MAJOR1 <= major && major <= I2O_MAJOR8);
 }
 
+static int
+_is_sx8_major (int major)
+{
+        return (SX8_MAJOR1 <= major && major <= SX8_MAJOR2);
+}
+
+#ifdef ENABLE_DEVICE_MAPPER
+static int
+readFD (int fd, char **buf)
+{
+        char* p;
+        size_t size = PROC_DEVICES_BUFSIZ;
+        int s, filesize = 0;
+
+        *buf = malloc (size * sizeof (char));
+        if (*buf == 0) {
+                return -1;
+        }
+
+        do {
+                p = &(*buf) [filesize];
+                s = read (fd, p, PROC_DEVICES_BUFSIZ);
+                /* exit if there is an error or EOF is reached */
+                if (s <= 0)
+                        break;
+                filesize += s;
+                size += s;
+                *buf = realloc (*buf, size);
+        } while (1);
+
+        if (filesize == 0 && s < 0) {
+                free (*buf);
+                *buf = NULL;
+                return -1;
+        } else {
+                /* there is always some excess memory left unused */
+                *buf = realloc (*buf, filesize+1);
+                (*buf)[filesize] = '\0';
+        }
+
+        return filesize;
+}
+
+static int
+_is_dm_major (int major)
+{
+        int fd;
+        char* buf = NULL;
+        char* line;
+        char* end;
+        int bd = 0;
+        int rc = 0;
+        char c;
+
+        fd = open ("/proc/devices", O_RDONLY);
+        if (fd < 0)
+                return 0;
+
+        if (readFD(fd, &buf) < 0) {
+                close(fd);
+                return 0;
+        }
+
+        line = buf;
+        end = strchr(line, '\n');
+        while (end) {
+                char *name;
+                int maj;
+
+                c = *end;
+                *end = '\0';
+
+                if (!bd) {
+                        if (!strncmp(line, "Block devices:", 14))
+                                bd = 1;
+                        goto next;
+                }
+
+                name = strrchr(line, ' ');
+                if (!name || strcmp(name+1, "device-mapper"))
+                        goto next;
+
+                maj = strtol(line, &name, 10);
+                if (maj == major) {
+                        free(buf);
+                        close(fd);
+                        return 1;
+                }
+
+next:
+                *end = c;
+                line = end+1;
+                end = strchr(line, '\n');
+        }
+        free(buf);
+        close(fd);
+        return 0;
+}
+#endif
+
 static int
 _device_stat (PedDevice* dev, struct stat * dev_stat)
 {
@@ -317,12 +452,22 @@ _device_probe_type (PedDevice* dev)
                 dev->type = PED_DEVICE_DAC960;
         } else if (dev_major == ATARAID_MAJOR && (dev_minor % 0x10 == 0)) {
                 dev->type = PED_DEVICE_ATARAID;
+        } else if (dev_major == DASD_MAJOR && (dev_minor % 0x4 == 0)) {
+                dev->type = PED_DEVICE_DASD;
+        } else if (dev_major == VIODASD_MAJOR && (dev_minor % 0x8 == 0)) {
+                dev->type = PED_DEVICE_VIODASD;
+        } else if (_is_sx8_major(dev_major) && (dev_minor % 0x20 == 0)) {
+                dev->type = PED_DEVICE_SX8;
         } else if (_is_i2o_major (dev_major) && (dev_minor % 0x10 == 0)) {
                 dev->type = PED_DEVICE_I2O;
         } else if (_is_cpqarray_major (dev_major) && (dev_minor % 0x10 == 0)) {
                 dev->type = PED_DEVICE_CPQARRAY;
         } else if (dev_major == UBD_MAJOR && (dev_minor % 0x10 == 0)) {
                 dev->type = PED_DEVICE_UBD;
+#ifdef ENABLE_DEVICE_MAPPER
+        } else if (_is_dm_major(dev_major)) {
+                dev->type = PED_DEVICE_DM;
+#endif
         } else {
                 dev->type = PED_DEVICE_UNKNOWN;
         }
@@ -394,6 +539,11 @@ _device_set_sector_size (PedDevice* dev)
                 dev->sector_size = (long long)sector_size;
         }
 
+        /* Return PED_SECTOR_SIZE_DEFAULT for DASDs. */
+        if (dev->type == PED_DEVICE_DASD) {
+                dev->sector_size = PED_SECTOR_SIZE_DEFAULT;
+        }
+
         if (dev->sector_size != PED_SECTOR_SIZE_DEFAULT) {
                 ped_exception_throw (
                         PED_EXCEPTION_WARNING,
@@ -803,6 +953,66 @@ error:
         return 0;
 }
 
+static int
+init_dasd (PedDevice* dev, char* model_name)
+{
+        struct stat             dev_stat;
+        PedExceptionOption      ex_status;
+        dasd_information_t      dasd_info;
+        struct hd_geometry      geo;
+        int f, blksize = 0;
+        char *errstr = 0;
+
+        if (!_device_stat (dev, &dev_stat))
+                goto error;
+
+        if (!ped_device_open (dev))
+                goto error;
+
+        LinuxSpecific* arch_specific = LINUX_SPECIFIC (dev);
+
+        PED_ASSERT (S_ISBLK (dev_stat.st_mode), return 0);
+
+        _device_set_sector_size (dev);
+        if (!dev->sector_size)
+                goto error_close_dev;
+
+        dev->length = _device_get_length (dev);
+        if (!dev->length)
+                goto error_close_dev;
+
+        if (!ioctl (arch_specific->fd, HDIO_GETGEO, &geo)) {
+                dev->hw_geom.sectors = geo.sectors;
+                dev->hw_geom.heads = geo.heads;
+                dev->hw_geom.cylinders = dev->length
+                        / (dev->hw_geom.heads * dev->hw_geom.sectors)
+                        / (dev->sector_size / PED_SECTOR_SIZE_DEFAULT);
+                dev->bios_geom = dev->hw_geom;
+        } else {
+                dev->bios_geom.sectors = 12;
+                dev->bios_geom.heads = 15;
+                dev->bios_geom.cylinders = dev->length
+                        / (dev->hw_geom.heads * dev->hw_geom.sectors)
+                        / (dev->sector_size / PED_SECTOR_SIZE_DEFAULT);
+                dev->hw_geom = dev->bios_geom;
+        }
+
+        dev->model = strdup (model_name);
+
+        ped_device_close (dev);
+        return 1;
+
+except:
+        ped_exception_throw ( PED_EXCEPTION_ERROR,
+                              PED_EXCEPTION_IGNORE_CANCEL,
+                              errstr );
+
+error_close_dev:
+        ped_device_close (dev);
+error:
+        return 0;
+}
+
 static int
 init_generic (PedDevice* dev, char* model_name)
 {
@@ -827,8 +1037,9 @@ init_generic (PedDevice* dev, char* model_name)
                                 PED_EXCEPTION_WARNING,
                                 PED_EXCEPTION_IGNORE_CANCEL,
                                 _("Unable to determine geometry of "
-                                "file/device.  You should not use Parted "
-                                "unless you REALLY know what you're doing!"));
+                                "file/device %s.  You should not use Parted "
+                                "unless you REALLY know what you're doing!"),
+                                dev->path);
                 switch (ex_status) {
                         case PED_EXCEPTION_CANCEL:
                                 goto error_close_dev;
@@ -904,6 +1115,21 @@ linux_new (const char* path)
                         goto error_free_arch_specific;
                 break;
 
+        case PED_DEVICE_SX8:
+                if (!init_generic (dev, _("Promise SX8 SATA Device")))
+                        goto error_free_arch_specific;
+                break;
+
+        case PED_DEVICE_DASD:
+                if (!init_dasd (dev, _("IBM S390 DASD drive")))
+                        goto error_free_arch_specific;
+                break;
+
+        case PED_DEVICE_VIODASD:
+                if (!init_generic (dev, _("IBM iSeries Virtual DASD")))
+                        goto error_free_arch_specific;
+                break;
+
         case PED_DEVICE_CPQARRAY:
                 if (!init_generic (dev, _("Compaq Smart Array")))
                         goto error_free_arch_specific;
@@ -921,7 +1147,7 @@ linux_new (const char* path)
 
         case PED_DEVICE_UBD:
                 if (!init_generic (dev, _("User-Mode Linux UBD")))
-                        goto error_free_dev;
+                        goto error_free_arch_specific;
                 break;
 
         case PED_DEVICE_FILE:
@@ -929,6 +1155,13 @@ linux_new (const char* path)
                         goto error_free_arch_specific;
                 break;
 
+#ifdef ENABLE_DEVICE_MAPPER
+        case PED_DEVICE_DM:
+                if (!init_generic (dev, _("Linux device-mapper")))
+                        goto error_free_arch_specific;
+                break;
+#endif
+
         case PED_DEVICE_UNKNOWN:
                 if (!init_generic (dev, _("Unknown")))
                         goto error_free_arch_specific;
@@ -1013,7 +1246,7 @@ _flush_cache (PedDevice* dev)
                 if (!name)
                         break;
                 if (!_partition_is_mounted_by_path (name)) {
-                        fd = open (name, O_WRONLY, 0);
+                        fd = open (name, WR_MODE, 0);
                         if (fd > 0) {
                                 ioctl (fd, BLKFLSBUF);
                                 close (fd);
@@ -1029,11 +1262,13 @@ linux_open (PedDevice* dev)
         LinuxSpecific*  arch_specific = LINUX_SPECIFIC (dev);
 
 retry:
-        arch_specific->fd = open (dev->path, O_RDWR);
+        arch_specific->fd = open (dev->path, RW_MODE);
+
         if (arch_specific->fd == -1) {
                 char*   rw_error_msg = strerror (errno);
 
-                arch_specific->fd = open (dev->path, O_RDONLY);
+                arch_specific->fd = open (dev->path, RD_MODE);
+
                 if (arch_specific->fd == -1) {
                         if (ped_exception_throw (
                                 PED_EXCEPTION_ERROR,
@@ -1170,12 +1405,14 @@ retry:
 }
 
 static int
-linux_read (const PedDevice* dev, void* buffer, PedSector start, PedSector count)
+linux_read (const PedDevice* dev, void* buffer, PedSector start,
+            PedSector count)
 {
         LinuxSpecific*          arch_specific = LINUX_SPECIFIC (dev);
         int                     status;
         PedExceptionOption      ex_status;
         size_t                  read_length = count * dev->sector_size;
+        void*                   diobuf;
 
         PED_ASSERT (dev->sector_size % 512 == 0, return 0);
 
@@ -1213,8 +1450,15 @@ linux_read (const PedDevice* dev, void* buffer, PedSector start, PedSector count
                 }
         }
 
+
+        if (posix_memalign(&diobuf, PED_SECTOR_SIZE_DEFAULT,
+                           count * PED_SECTOR_SIZE_DEFAULT) != 0)
+                return 0;
+
         while (1) {
-                status = read (arch_specific->fd, buffer, read_length);
+                status = read (arch_specific->fd, diobuf, read_length);
+                if (status > 0)
+                        memcpy(buffer, diobuf, status);
                 if (status == count * dev->sector_size) break;
                 if (status > 0) {
                         read_length -= status;
@@ -1231,6 +1475,7 @@ linux_read (const PedDevice* dev, void* buffer, PedSector start, PedSector count
 
                 switch (ex_status) {
                         case PED_EXCEPTION_IGNORE:
+                                free(diobuf);
                                 return 1;
 
                         case PED_EXCEPTION_RETRY:
@@ -1239,9 +1484,11 @@ linux_read (const PedDevice* dev, void* buffer, PedSector start, PedSector count
                         case PED_EXCEPTION_UNHANDLED:
                                 ped_exception_catch ();
                         case PED_EXCEPTION_CANCEL:
+                                free(diobuf);
                                 return 0;
                 }
         }
+        free(diobuf);
 
         return 1;
 }
@@ -1287,6 +1534,8 @@ linux_write (PedDevice* dev, const void* buffer, PedSector start,
         int                     status;
         PedExceptionOption      ex_status;
         size_t                  write_length = count * dev->sector_size;
+        void*                   diobuf;
+        void*                   diobuf_start;
 
         PED_ASSERT(dev->sector_size % 512 == 0, return 0);
 
@@ -1340,12 +1589,17 @@ linux_write (PedDevice* dev, const void* buffer, PedSector start,
                 dev->path, buffer, (int) start, (int) count);
 #else
         dev->dirty = 1;
+        if (posix_memalign(&diobuf, PED_SECTOR_SIZE_DEFAULT,
+                           count * PED_SECTOR_SIZE_DEFAULT) != 0)
+                return 0;
+        memcpy(diobuf, buffer, count * PED_SECTOR_SIZE_DEFAULT);
+        diobuf_start = diobuf;
         while (1) {
-                status = write (arch_specific->fd, buffer, write_length);
+                status = write (arch_specific->fd, diobuf, write_length);
                 if (status == count * dev->sector_size) break;
                 if (status > 0) {
                         write_length -= status;
-                        buffer += status;
+                        diobuf += status;
                         continue;
                 }
 
@@ -1357,6 +1611,7 @@ linux_write (PedDevice* dev, const void* buffer, PedSector start,
 
                 switch (ex_status) {
                         case PED_EXCEPTION_IGNORE:
+                                free(diobuf_start);
                                 return 1;
 
                         case PED_EXCEPTION_RETRY:
@@ -1365,9 +1620,11 @@ linux_write (PedDevice* dev, const void* buffer, PedSector start,
                         case PED_EXCEPTION_UNHANDLED:
                                 ped_exception_catch ();
                         case PED_EXCEPTION_CANCEL:
+                                free(diobuf_start);
                                 return 0;
                 }
         }
+        free(diobuf_start);
 #endif /* !READ_ONLY */
         return 1;
 }
@@ -1380,18 +1637,26 @@ linux_check (PedDevice* dev, void* buffer, PedSector start, PedSector count)
         LinuxSpecific*  arch_specific = LINUX_SPECIFIC (dev);
         PedSector       done = 0;
         int             status;
+        void*           diobuf;
 
         PED_ASSERT(dev != NULL, return 0);
         
         if (!_device_seek (dev, start))
                 return 0;
 
+        if (posix_memalign(&diobuf, PED_SECTOR_SIZE_DEFAULT,
+                           count * PED_SECTOR_SIZE_DEFAULT) != 0)
+                return 0;
+
         for (done = 0; done < count; done += status / dev->sector_size) {
-                status = read (arch_specific->fd, buffer,
+                status = read (arch_specific->fd, diobuf,
                                (size_t) ((count-done) * dev->sector_size));
+                if (status > 0)
+                        memcpy(buffer, diobuf, status);
                 if (status < 0)
                         break;
         }
+        free(diobuf);
 
         return done;
 }
@@ -1647,6 +1912,9 @@ _device_get_part_path (PedDevice* dev, int num)
         } else if (dev->type == PED_DEVICE_DAC960
                         || dev->type == PED_DEVICE_CPQARRAY
                         || dev->type == PED_DEVICE_ATARAID
+#ifdef ENABLE_DEVICE_MAPPER
+                        || dev->type == PED_DEVICE_DM
+#endif
                         || isdigit (dev->path[path_len - 1]))
                 snprintf (result, result_len, "%sp%d", dev->path, num);
         else
@@ -1782,7 +2050,11 @@ _blkpg_add_partition (PedDisk* disk, PedPartition* part)
 
         memset (&linux_part, 0, sizeof (linux_part));
         linux_part.start = part->geom.start * disk->dev->sector_size;
-        linux_part.length = part->geom.length * disk->dev->sector_size;
+        /* see fs/partitions/msdos.c:msdos_partition(): "leave room for LILO" */
+        if (part->type & PED_PARTITION_EXTENDED)
+                linux_part.length = part->geom.length == 1 ? 512 : 1024;
+        else
+                linux_part.length = part->geom.length * disk->dev->sector_size;
         linux_part.pno = part->num;
         strncpy (linux_part.devname, dev_name, BLKPG_DEVNAMELTH);
         if (vol_name)
@@ -1839,13 +2111,6 @@ _disk_sync_part_table (PedDisk* disk)
 
                 part = ped_disk_get_partition (disk, i);
                 if (part) {
-                        /* extended partitions have no business in the kernel!
-                         * blkpg doesn't like overlapping partitions.  Hmmm,
-                         * LILO isn't going to like this.
-                         */
-                        if (part->type & PED_PARTITION_EXTENDED)
-                                continue;
-
                         /* busy... so we won't (can't!) disturb ;)  Prolly
                          * doesn't matter anyway, because users shouldn't be
                          * changing mounted partitions anyway...
@@ -1862,6 +2127,239 @@ _disk_sync_part_table (PedDisk* disk)
         return ret;
 }
 
+#ifdef ENABLE_DEVICE_MAPPER
+static int
+_dm_remove_map(int major, int minor)
+{
+        struct dm_task  *task = NULL;
+        int             rc;
+
+        task = dm_task_create(DM_DEVICE_REMOVE);
+        if (!task)
+                return 1;
+
+        dm_task_set_major (task, major);
+        dm_task_set_minor (task, minor);
+
+        rc = dm_task_run(task);
+        dm_task_update_nodes();
+        dm_task_destroy(task);
+        if (rc < 0)
+                return 1;
+
+        return 0;
+}
+
+static int
+_dm_remove_map_name(char *name)
+{
+        struct dm_task  *task = NULL;
+        int             rc;
+
+        task = dm_task_create(DM_DEVICE_REMOVE);
+        if (!task)
+                return 1;
+
+        dm_task_set_name (task, name);
+
+        rc = dm_task_run(task);
+        dm_task_update_nodes();
+        dm_task_destroy(task);
+        if (rc < 0)
+                return 1;
+
+        return 0;
+}
+
+static int 
+_dm_is_part (struct dm_info *this, char *name)
+{
+        struct dm_task* task = NULL;
+        struct dm_info* info = alloca(sizeof *info);
+        struct dm_deps* deps = NULL;
+        int             rc = 0;
+        unsigned int    i;
+
+        task = dm_task_create(DM_DEVICE_DEPS);
+        if (!task)
+                return 0;
+        
+        dm_task_set_name(task, name);
+        rc = dm_task_run(task);
+        if (rc < 0) {
+                rc = 0;
+                goto err;
+        }
+        rc = 0;
+
+        memset(info, '\0', sizeof *info);
+        dm_task_get_info(task, info);
+        if (!info->exists)
+                goto err;
+
+        deps = dm_task_get_deps(task);
+        if (!deps)
+                goto err;
+
+        rc = 0;
+        for (i = 0; i < deps->count; i++) {
+                unsigned int ma = major(deps->device[i]),
+                             mi = minor(deps->device[i]);
+
+                if (ma == this->major && mi == this->minor)
+                        rc = 1;
+        }
+
+err:
+        dm_task_destroy(task);
+        return rc;
+}
+
+static int
+_dm_remove_parts (PedDevice* dev)
+{
+        struct stat             dev_stat;
+        struct dm_task*         task = NULL;
+        struct dm_info*         info = alloca(sizeof *info);
+        struct dm_names*        names = NULL;
+        unsigned int            next = 0;
+        int                     i;
+        int                     rc;
+
+        if (!_device_stat (dev, &dev_stat))
+                goto err;
+
+        task = dm_task_create(DM_DEVICE_LIST);
+        if (!task)
+                goto err;
+
+        dm_task_set_major (task, major (dev_stat.st_rdev));
+        dm_task_set_minor (task, minor (dev_stat.st_rdev));
+
+        rc = dm_task_run(task);
+        if (rc < 0)
+                goto err;
+
+        memset(info, '\0', sizeof *info);
+        dm_task_get_info(task, info);
+        if (!info->exists)
+                goto err;
+
+        names = dm_task_get_names(task);
+        if (!names)
+                goto err;
+
+        rc = 0;
+        do {
+                names = (void *)names + next;
+
+                if (_dm_is_part(info, names->name))
+                        rc += _dm_remove_map_name(names->name);
+
+                next = names->next;
+        } while (next);
+
+        dm_task_update_nodes();
+        dm_task_destroy(task);
+        task = NULL;
+
+        if (!rc)
+                return 1;
+err:
+        if (task)
+                dm_task_destroy(task);
+        ped_exception_throw (PED_EXCEPTION_WARNING, PED_EXCEPTION_IGNORE,
+                _("parted was unable to re-read the partition "
+                  "table on %s (%s).  This means Linux won't know "
+                  "anything about the modifications you made. "),
+                dev->path, strerror (errno));
+        return 0;
+}
+
+static int
+_dm_add_partition (PedDisk* disk, PedPartition* part)
+{
+        struct dm_task* task;
+        struct stat     dev_stat;
+        int             rc;
+        char*           vol_name = NULL;
+        char*           dev_name = NULL;
+        char*           params = NULL;
+
+        dev_name = _device_get_part_path (disk->dev, part->num);
+        if (!dev_name)
+                return 0;
+
+        vol_name = strrchr (dev_name, '/');
+        if (vol_name && *vol_name && *(++vol_name))
+                vol_name = strdup (vol_name);
+        else
+                vol_name = strdup (dev_name);
+        if (!vol_name)
+                return 0;
+
+        if (!_device_stat (disk->dev, &dev_stat))
+                goto err;
+
+        asprintf (&params, "%d:%d %lu", major (dev_stat.st_rdev),
+                        minor (dev_stat.st_rdev),
+                        part->geom.start);
+        if (!params)
+                goto err;
+
+        task = dm_task_create (DM_DEVICE_CREATE);
+        if (!task)
+                goto err;
+
+        dm_task_set_name (task, vol_name);
+        dm_task_add_target (task, 0, part->geom.length,
+                "linear", params);
+        rc = dm_task_run(task);
+        if (rc >= 0) {
+                //printf("0 %ld linear %s\n", part->geom.length, params);
+                dm_task_update_nodes();
+                dm_task_destroy(task);
+                free(params);
+                free(vol_name);
+                return 1;
+        } else {
+                _dm_remove_map_name(vol_name);
+        }
+err:
+        dm_task_update_nodes();
+        if (task)
+                dm_task_destroy (task);
+        if (params)
+                free (params);
+        free (vol_name);
+        return 0;
+}
+
+static int
+_dm_reread_part_table (PedDisk* disk)
+{
+        int     rc = 1;
+        int     last = PED_MAX (ped_disk_get_last_partition_num (disk), 16);
+        int     i;
+
+        sync();
+        if (!_dm_remove_parts(disk->dev))
+                rc = 0;
+
+        for (i = 1; i <= last; i++) {
+                PedPartition*      part;
+
+                part = ped_disk_get_partition (disk, i);
+                if (!part)
+                        continue;
+
+                if (!_dm_add_partition (disk, part))
+                        rc = 0;
+        }
+        return rc;
+}
+#endif
+
 static int
 _kernel_reread_part_table (PedDevice* dev)
 {
@@ -1905,6 +2403,10 @@ _have_blkpg ()
 static int
 linux_disk_commit (PedDisk* disk)
 {
+#ifdef ENABLE_DEVICE_MAPPER
+        if (disk->dev->type == PED_DEVICE_DM)
+                return _dm_reread_part_table (disk);
+#endif
         if (disk->dev->type != PED_DEVICE_FILE) {
                 /* The ioctl() command BLKPG_ADD_PARTITION does not notify
                  * the devfs system; consequently, /proc/partitions will not
@@ -1912,12 +2414,16 @@ linux_disk_commit (PedDisk* disk)
                  * created.  Therefore, if using DevFS, we must get the kernel
                  * to re-read and grok the partition table.
                  */
-                if (_have_blkpg () && !_have_devfs ()) {
+                /* Work around kernel dasd problem so we really do BLKRRPART */
+                if (disk->dev->type != PED_DEVICE_DASD &&
+                    _have_blkpg () && !_have_devfs ()) {
                         if (_disk_sync_part_table (disk))
                                 return 1;
                 }
+
                 return _kernel_reread_part_table (disk->dev);
         }
+
         return 1;
 }
 
