@@ -166,7 +166,9 @@ _amiga_calculate_checksum (struct AmigaBlock *blk) {
 }
 
 static struct AmigaBlock *
-_amiga_read_block (const PedDevice *dev, struct AmigaBlock *blk, PedSector block, struct AmigaIds *ids) {
+_amiga_read_block (const PedDevice *dev, struct AmigaBlock *blk,
+                   PedSector block, struct AmigaIds *ids)
+{
 	if (!ped_device_read (dev, blk, block, 1))
 		return NULL;
 	if (ids && !_amiga_id_in_list(PED_BE32_TO_CPU(blk->amiga_ID), ids))
@@ -276,10 +278,12 @@ struct PartitionBlock {
     char	pb_DriveName[32];	/* Preferred DOS device name: BSTR form */
     uint32_t	pb_Reserved2[15];
     uint32_t	de_TableSize;		/* Size of Environment vector */
-    uint32_t	de_SizeBlock;		/* Size of the blocks in 32 bit words, usually 128 */
+	/* Size of the blocks in 32 bit words, usually 128 */
+    uint32_t	de_SizeBlock;
     uint32_t	de_SecOrg;	     	/* Not used; must be 0 */
     uint32_t	de_Surfaces;		/* Number of heads (surfaces) */
-    uint32_t	de_SectorPerBlock;	/* Disk sectors per block, used with SizeBlock, usually 1 */
+	/* Disk sectors per block, used with SizeBlock, usually 1 */
+    uint32_t	de_SectorPerBlock;
     uint32_t	de_BlocksPerTrack;	/* Blocks per track. drive specific */
     uint32_t	de_Reserved;		/* DOS reserved blocks at start of partition. */
     uint32_t	de_PreAlloc;		/* DOS reserved blocks at end of partition */
@@ -545,7 +549,8 @@ amiga_read (PedDisk* disk)
 		}
 
 		/* Let's allocate and read a partition block to get its geometry*/
-		if (!_amiga_read_block (disk->dev, AMIGA(partition), (PedSector)partblock, NULL)) {
+		if (!_amiga_read_block (disk->dev, AMIGA(partition),
+		                        (PedSector)partblock, NULL)) {
 			ped_free(partition);
 			return 0;
 		}
@@ -728,7 +733,8 @@ amiga_write (PedDisk* disk)
 		goto error_free_table;
 	}
 
-	block_num = next_num = part_num = _amiga_next_free_block(table, rdb_num+1, IDNAME_PARTITION);
+	block_num = next_num = part_num = _amiga_next_free_block(table, rdb_num+1,
+	                                                         IDNAME_PARTITION);
 	part = _amiga_next_real_partition(disk, NULL);
 	rdb->rdb_PartitionList = PED_CPU_TO_BE32(part ? part_num : LINK_END);
 	for (; part != NULL; part = next_part, block_num = next_num) {
@@ -740,7 +746,10 @@ amiga_write (PedDisk* disk)
 		next_num = _amiga_next_free_block(table, block_num+1, IDNAME_PARTITION);
 
 		partition = PART(part->disk_specific);
-		partition->pb_Next = next_part==NULL ? PED_CPU_TO_BE32(LINK_END):PED_CPU_TO_BE32(next_num);
+		if (next_part == NULL)
+			partition->pb_Next = PED_CPU_TO_BE32(LINK_END);
+		else
+			partition->pb_Next = PED_CPU_TO_BE32(next_num);
 		partition->de_LowCyl = PED_CPU_TO_BE32(part->geom.start/cylblocks);
 		partition->de_HighCyl = PED_CPU_TO_BE32((part->geom.end+1)/cylblocks-1);
 		_amiga_calculate_checksum(AMIGA(partition));
@@ -881,7 +890,8 @@ amiga_partition_destroy (PedPartition* part)
 }
 
 static int
-amiga_partition_set_system (PedPartition* part, const PedFileSystemType* fs_type)
+amiga_partition_set_system (PedPartition* part,
+                            const PedFileSystemType* fs_type)
 {
 	struct PartitionBlock *partition;
 
@@ -1096,7 +1106,8 @@ amiga_alloc_metadata (PedDisk* disk)
 	constraint_any = ped_constraint_any (disk->dev);
 
 	/* Allocate space for the RDB */
-	new_part = ped_partition_new (disk, PED_PARTITION_METADATA, NULL, 0, MAX_RDB_BLOCK);
+	new_part = ped_partition_new (disk, PED_PARTITION_METADATA, NULL,
+	                              0, MAX_RDB_BLOCK);
 	if (!new_part)
 		goto error;
 
